@@ -2,20 +2,36 @@ import hashlib
 import logging
 import os
 import random
+from enum import Enum
 from pathlib import Path
-from typing import Any, Callable, Dict, Optional
+from typing import Any, Callable, Dict, List, Optional, Tuple
 
 import cloudpickle
 import colorlog
 import redis
+import yaml
 from pydantic import BaseModel
 
-from motion.dicts import CustomDict, State
+from motion.dicts import State
 
 logger = logging.getLogger(__name__)
 
 DEFAULT_KEY_TTL = 60 * 60 * 24  # 1 day
-fake_custom_dict = CustomDict("fake", "fake", "fake", {})
+
+
+def import_config(config_path: str = ".motionrc.yml") -> None:
+    # If env var MOTION_YAML_LOADED is not set, load .motionrc.yml
+    if os.getenv("MOTION_YAML_LOADED") is None:
+        # Load .motionrc.yml if it exists
+        if os.path.exists(config_path):
+            with open(config_path, "r") as f:
+                config = yaml.safe_load(f)
+
+            # Set env vars to the values in .motionrc.yml
+            for k, v in config.items():
+                os.environ[k] = str(v)
+
+        os.environ["MOTION_YAML_LOADED"] = "1"
 
 
 def hash_object(obj: Any) -> str:
@@ -31,21 +47,91 @@ def hash_object(obj: Any) -> str:
     return hex_digest
 
 
-class RedisParams(BaseModel):
+class RedisParams(BaseModel, extra="allow"):
     host: str
     port: int
     db: int
     password: Optional[str] = None
+    ssl: bool = False
 
     def __init__(self, **kwargs: Any) -> None:
         kwargs.setdefault("host", os.getenv("MOTION_REDIS_HOST", "localhost"))
         kwargs.setdefault("port", int(os.getenv("MOTION_REDIS_PORT", "6379")))
         kwargs.setdefault("db", int(os.getenv("MOTION_REDIS_DB", "0")))
-        kwargs.setdefault("password", os.getenv("MOTION_REDIS_PASSWORD", None))
+
+        if str(os.getenv("MOTION_REDIS_PASSWORD", "None")) != "None":
+            kwargs["password"] = os.getenv("MOTION_REDIS_PASSWORD")
+
+        if str(os.getenv("MOTION_REDIS_SSL", "False")) == "True":
+            kwargs["ssl"] = True
 
         super().__init__(**kwargs)
 
 
+def get_redis_params() -> RedisParams:
+    import_config()
+    rp = RedisParams()
+    return rp
+
+
+def get_instances(component_name: str) -> List[str]:
+    """Gets all instances of a component.
+
+    Args:
+        component_name (str): Name of the component.
+
+    Returns:
+        List[str]: List of instance ids.
+    """
+    rp = get_redis_params()
+    redis_con = redis.Redis(**rp.dict())
+
+    instance_ids = redis_con.keys(f"MOTION_VERSION:{component_name}__*")
+    instance_ids = [
+        instance_id.decode("utf-8").split("__")[1] for instance_id in instance_ids
+    ]
+
+    redis_con.close()
+
+    return instance_ids
+
+
+def clear_dev_instances() -> int:
+    """Clears all dev instances."""
+    rp = get_redis_params()
+    redis_con = redis.Redis(**rp.dict())
+
+    # Scan for all keys with prefix
+    prefix = "MOTION_VERSION:DEV:*"
+    pipeline = redis_con.pipeline()
+    num_keys_deleted = 0
+    for key in redis_con.scan_iter(prefix):
+        pipeline.delete(key)
+        num_keys_deleted += 1
+
+    # Delete all states too
+    prefix = "MOTION_STATE:DEV:*"
+    for key in redis_con.scan_iter(prefix):
+        pipeline.delete(key)
+
+    results_to_delete = redis_con.keys("MOTION_RESULT:DEV:*")
+    queues_to_delete = redis_con.keys("MOTION_QUEUE:DEV:*")
+    locks_to_delete = redis_con.keys("MOTION_LOCK:DEV:*")
+    for result in results_to_delete:
+        pipeline.delete(result)
+    for queue in queues_to_delete:
+        pipeline.delete(queue)
+    for lock in locks_to_delete:
+        pipeline.delete(lock)
+
+    pipeline.execute()
+    pipeline.close()
+
+    redis_con.close()
+
+    return num_keys_deleted
+
+
 def clear_instance(instance_name: str) -> bool:
     """Clears the state and cached results associated with a component instance.
 
@@ -71,30 +157,46 @@ def clear_instance(instance_name: str) -> bool:
     if "__" not in instance_name:
         raise ValueError("Instance must be in the form `componentname__instanceid`.")
 
-    rp = RedisParams()
-    redis_con = redis.Redis(host=rp.host, port=rp.port, password=rp.password, db=rp.db)
+    rp = get_redis_params()
+    redis_con = redis.Redis(
+        **rp.dict(),
+    )
 
     # Check if the instance exists
-    if not redis_con.exists(f"MOTION_VERSION:{instance_name}"):
+    if not redis_con.exists(f"MOTION_VERSION:{instance_name}") and not redis_con.exists(
+        f"MOTION_VERSION:DEV:{instance_name}"
+    ):
         return False
 
     # Delete the instance state, version, and cached results
     redis_con.delete(f"MOTION_STATE:{instance_name}")
+    redis_con.delete(f"MOTION_STATE:DEV:{instance_name}")
+    redis_con.delete(f"MOTION_VERSION:DEV:{instance_name}")
     redis_con.delete(f"MOTION_VERSION:{instance_name}")
+    redis_con.delete(f"MOTION_LOCK:{instance_name}")
+    redis_con.delete(f"MOTION_LOCK:DEV:{instance_name}")
+
+    for env in [":DEV", ""]:
+        results_to_delete = redis_con.keys(f"MOTION_RESULT{env}:{instance_name}/*")
+        queues_to_delete = redis_con.keys(f"MOTION_QUEUE{env}:{instance_name}/*")
+        channels_to_delete = redis_con.keys(f"MOTION_CHANNEL{env}:{instance_name}/*")
+
+        pipeline = redis_con.pipeline()
+        for result in results_to_delete:
+            pipeline.delete(result)
+        for queue in queues_to_delete:
+            pipeline.delete(queue)
+        for channel in channels_to_delete:
+            pipeline.delete(channel)
 
-    results_to_delete = redis_con.keys(f"MOTION_RESULT:{instance_name}/*")
-    queues_to_delete = redis_con.keys(f"MOTION_QUEUE:{instance_name}/*")
-    pipeline = redis_con.pipeline()
-    for result in results_to_delete:
-        pipeline.delete(result)
-    for queue in queues_to_delete:
-        pipeline.delete(queue)
     pipeline.execute()
 
+    redis_con.close()
+
     return True
 
 
-def inspect_state(instance_name: str) -> Dict[str, Any]:
+def inspect_state(instance_name: str) -> Optional[State]:
     """
     Returns the state of a component instance.
 
@@ -115,23 +217,48 @@ def inspect_state(instance_name: str) -> Dict[str, Any]:
             `componentname__instanceid` or if the instance does not exist.
 
     Returns:
-        Dict[str, Any]: The state of the component instance.
+        State: The state of the component instance.
     """
     if "__" not in instance_name:
         raise ValueError("Instance must be in the form `componentname__instanceid`.")
 
-    rp = RedisParams()
-    redis_con = redis.Redis(host=rp.host, port=rp.port, password=rp.password, db=rp.db)
+    rp = get_redis_params()
+    redis_con = redis.Redis(
+        **rp.dict(),
+    )
 
     # Check if the instance exists
     if not redis_con.exists(f"MOTION_VERSION:{instance_name}"):
         raise ValueError(f"Instance {instance_name} does not exist.")
 
     # Get the state
-    state = loadState(redis_con, instance_name, None)
+    state, _ = loadState(redis_con, instance_name, None)
+
+    redis_con.close()
     return state
 
 
+def get_components() -> List[str]:
+    """Lists all components in the Redis database.
+
+    Returns:
+        List[str]: List of all component names.
+    """
+
+    rp = get_redis_params()
+    redis_con = redis.Redis(
+        **rp.dict(),
+    )
+
+    component_names = [
+        name.decode("utf-8") for name in redis_con.smembers("MOTION_COMPONENTS")
+    ]
+
+    redis_con.close()
+
+    return component_names
+
+
 def validate_args(parameters: Any, op: str) -> bool:
     if "state" not in parameters.keys():
         return False
@@ -146,10 +273,9 @@ def validate_args(parameters: Any, op: str) -> bool:
 
 
 def configureLogging(level: str) -> None:
-    handler = logging.StreamHandler()
-
     formatter = colorlog.ColoredFormatter(
         "%(log_color)s%(asctime)s %(levelname)-8s%(reset)s %(blue)s%(message)s",
+        datefmt="%Y-%m-%d %H:%M:%S",
         log_colors={
             "DEBUG": "cyan",
             "INFO": "green",
@@ -158,9 +284,15 @@ def configureLogging(level: str) -> None:
             "CRITICAL": "bold_red",
         },
     )
-    handler.setFormatter(formatter)
+
     logger = logging.getLogger("motion")
-    logger.addHandler(handler)
+    if logger.hasHandlers():
+        logger.handlers.clear()
+
+    stream_handler = logging.StreamHandler()
+    stream_handler.setFormatter(formatter)
+
+    logger.addHandler(stream_handler)
     logger.setLevel(level)
 
 
@@ -168,15 +300,31 @@ def loadState(
     redis_con: redis.Redis,
     instance_name: str,
     load_state_func: Optional[Callable],
-) -> State:
+) -> Tuple[Optional[State], int]:
     # Get state from redis
     state = State(instance_name.split("__")[0], instance_name.split("__")[1], {})
-    loaded_state = redis_con.get(f"MOTION_STATE:{instance_name}")
+
+    # If dev mode, load with diff prefix
+    loaded_state = None
+    version = None
+    if os.getenv("MOTION_ENV", "prod") == "dev":
+        loaded_state = redis_con.get(f"MOTION_STATE:DEV:{instance_name}")
+        if loaded_state:
+            v_identifier = f"MOTION_VERSION:DEV:{instance_name}"
+            version = int(redis_con.get(v_identifier))  # type: ignore
+        else:
+            loaded_state = redis_con.get(f"MOTION_STATE:{instance_name}")
+
+    else:
+        loaded_state = redis_con.get(f"MOTION_STATE:{instance_name}")
 
     if not loaded_state:
         # This is an error
-        logger.warning(f"Could not find state for {instance_name}.")
-        return state
+        logger.warning(f"Could not find state for {instance_name}. Creating new state.")
+        return None, 0
+
+    if not version:
+        version = int(redis_con.get(f"MOTION_VERSION:{instance_name}"))  # type: ignore
 
     # Unpickle state
     loaded_state = cloudpickle.loads(loaded_state)
@@ -186,23 +334,44 @@ def loadState(
     else:
         state.update(loaded_state)
 
-    return state
+    return state, version
 
 
 def saveState(
     state_to_save: State,
+    version: int,
     redis_con: redis.Redis,
     instance_name: str,
     save_state_func: Optional[Callable],
-) -> None:
+) -> int:
+    # If the version in redis is greater than this version, drop the save
+    redis_v = None
+    if os.getenv("MOTION_ENV", "prod") == "dev":
+        redis_con.get(f"MOTION_VERSION:DEV:{instance_name}")
+
+    if not redis_v:
+        redis_v = redis_con.get(f"MOTION_VERSION:{instance_name}")
+
+    if redis_v and int(redis_v) > version:
+        # This means that another process has already saved the state
+        # Return a sentinel value that indicates that the state was not saved
+        return -1
+
     # Save state to redis
     if save_state_func is not None:
         state_to_save = save_state_func(state_to_save)
 
     state_pickled = cloudpickle.dumps(state_to_save)
 
-    redis_con.set(f"MOTION_STATE:{instance_name}", state_pickled)
-    redis_con.incr(f"MOTION_VERSION:{instance_name}")
+    if os.getenv("MOTION_ENV", "prod") == "dev":
+        redis_con.set(f"MOTION_STATE:DEV:{instance_name}", state_pickled)
+        redis_con.set(f"MOTION_VERSION:DEV:{instance_name}", version + 1)
+
+    else:
+        redis_con.set(f"MOTION_STATE:{instance_name}", state_pickled)
+        redis_con.set(f"MOTION_VERSION:{instance_name}", version + 1)
+
+    return version + 1
 
 
 class UpdateEvent:
@@ -249,7 +418,7 @@ class UpdateEventGroup:
         self.events[udf_name] = event
 
     def wait(self) -> None:
-        """Waits for all update operations for this dataflow key
+        """Waits for all update operations for this flow key
         to finish.
 
         Example usage:
@@ -304,3 +473,9 @@ def random_passphrase(num_words: int = 3) -> str:
     ]
     words = [wordlist[roll] for roll in dice_rolls]
     return "-".join(words)
+
+
+# Status Enum
+class FlowOpStatus(Enum):
+    SUCCESS = "Success"
+    FAILURE = "Failure"
